using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;

// Wall3DTilesEditor v1.0

namespace Wall3DTilesEditorPlugin
{
    public class Wall3DTilesEditor : MVRScript
    {
		// JSONStorables
		public JSONStorableBool ToggleEditTilesParameters;
		public JSONStorableBool ToggleAnimateTilesParameters;
		public JSONStorableBool ToggleMaterialParameters;
		
		public JSONStorableBool EnableCollisions;
		
		public JSONStorableStringChooser TileType;
		protected List<string> TileChoices;
		protected List<Transform> TileChoicesTRs;
		public JSONStorableStringChooser TileTypeAlt;
		protected List<string> TileAltChoices;
		
		public JSONStorableStringChooser TileMaterial;
		protected List<string> TileMaterialChoices;
		protected List<Transform> TileMaterialTRs;
		
		public JSONStorableFloat TileLineCount;
		public JSONStorableFloat TileRowCount;
		
		public JSONStorableStringChooser TileSpawnRadiusType;
		protected List<string> TileSpawnRadiusTypesChoice;
		
		public JSONStorableFloat TileSpawnRadius;
		public JSONStorableFloat TileSpawnCircumference;
		public JSONStorableFloat TileSpawnRadialSpacing;
		public JSONStorableFloat TileHorizontalSpacing;
		public JSONStorableFloat TileVerticalSpacing;
		public JSONStorableFloat TileLineHorizontalOffset;
		public JSONStorableFloat TileLineVerticalOffset;
		public JSONStorableFloat TileLineRotationOffset;
		public JSONStorableFloat TileScale;
		public JSONStorableFloat TileRotationX;
		public JSONStorableFloat TileRotationY;
		public JSONStorableFloat TileRotationZ;
		
		public JSONStorableColor MatAlbedoColor;
		public JSONStorableColor MatBaseColor;
		public JSONStorableFloat MatAlbedoBaseColorRatio;
		public JSONStorableFloat MatRoughnessIntensity;
		public JSONStorableFloat MatNormalIntensity;
		public JSONStorableFloat MatScale;
		
		private JSONStorableString _loadingInfos;
		private JSONStorableString _helpText;
		
		// JSONStorables (Animation only)
		public JSONStorableBool EnableAnimation;
		public JSONStorableColor AnimateAlbedoColor;
		public JSONStorableColor AnimateBaseColor;
		public JSONStorableFloat AnimateScale;
		public JSONStorableFloat AnimateRotationX;
		public JSONStorableFloat AnimateRotationY;
		public JSONStorableFloat AnimateRotationZ;
		
		// Tiles for the animation
		protected List<Transform> CurrentTilesList;
		protected Transform[] CurrentTilesArray;

		// Game Objects and components
		private Transform TilesObjectsRoot;
		private Transform TilesMaterialsRoot;
		private Transform TilesRoot;
		
		// Security and initialization
		private bool editorInitialized = false;
		private int initLoops = 0;
		private int maxInitLoops = 500;
		private bool cuaIsValid = false;
		private bool animRoutineRunning = false;
		
		// UI Elements		
		private UIDynamic HelpUIElement;
		private UIDynamic[] colorUIElements;
		
		private UIDynamic[] tilesParamsUIElements;
		private UIDynamic[] tilesAnimUIElements;
		private UIDynamic[] tilesMaterialUIElements;
		
		private List<UIDynamicColorPicker> allUIColorPickers;
		private List<UIDynamicSlider> allUISliders;
		private List<UIDynamicToggle> allUIToggles;
		private List<UIDynamicTextField> allUITextFields;
		private List<UIDynamicButton> allUIButtons;
		private List<UIDynamic> allUISpacers;
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add Wall3DTilesEditor on a Custom Unity Asset");
                    return;
                }

				ToggleEditTilesParameters = new JSONStorableBool("Show tiles parameters", true,ToggleEditTilesParametersCallback){ isStorable = false };
				ToggleAnimateTilesParameters = new JSONStorableBool("Show animation parameters", false,ToggleAnimateTilesParametersCallback){ isStorable = false };
				ToggleMaterialParameters = new JSONStorableBool("Show material parameters", false,ToggleMaterialParametersCallback){ isStorable = false };
				
				EnableCollisions = new JSONStorableBool("Enable collisions", false);
				EnableCollisions.setCallbackFunction += (val) => { EnableCollisions.valNoCallback = val; CreateTiles(); };
				
				TileChoices = new List<string>();
				
				TileType = new JSONStorableStringChooser("Tile type", TileChoices, "tile_01", "Tile type");
				TileType.setCallbackFunction += (val) => { TileType.valNoCallback = val; CreateTiles(); };
				
				TileAltChoices = new List<string>();
				TileAltChoices.Add("None");
				
				TileTypeAlt = new JSONStorableStringChooser("Tile type (alt)", TileAltChoices, "None", "Tile type (alt)");
				TileTypeAlt.setCallbackFunction += (val) => { TileTypeAlt.valNoCallback = val; CreateTiles(); };
				
				TileMaterial = new JSONStorableStringChooser("Tile material", TileMaterialChoices, "marble_01", "Tile material");
				TileMaterial.setCallbackFunction += (val) => { TileMaterial.valNoCallback = val; CreateTiles(); };
				
				TileLineCount = new JSONStorableFloat("Line count", 4f, 1f, 20f, false);
				TileLineCount.setCallbackFunction += (val) => { TileLineCount.valNoCallback = Mathf.Round(val); CreateTiles(); }; // This is meant to avoid creating a specific callback for all the storables and create the tiles immediately
				TileRowCount = new JSONStorableFloat("Row count", 10f, 1f, 50f, false);
				TileRowCount.setCallbackFunction += (val) => { TileRowCount.valNoCallback = Mathf.Round(val); CreateTiles(); };
				
				TileSpawnRadiusTypesChoice = new List<string>();
				TileSpawnRadiusTypesChoice.Add("Flat");
				TileSpawnRadiusTypesChoice.Add("Radial Plane");
				TileSpawnRadiusTypesChoice.Add("Radial 3D");
				
				TileSpawnRadiusType = new JSONStorableStringChooser("Spawn Radius type", TileSpawnRadiusTypesChoice, "Flat", "Spawn Radius type");
				TileSpawnRadiusType.setCallbackFunction += (val) =>
				{
					TileSpawnRadiusType.valNoCallback = val; CreateTiles();
					ToggleEditRadiusElements();
				};
				TileSpawnRadius = new JSONStorableFloat("Spawn Radius", 1f, 0.01f, 10f);
				TileSpawnRadius.setCallbackFunction += (val) => { TileSpawnRadius.valNoCallback = val; CreateTiles(); };
				TileSpawnCircumference = new JSONStorableFloat("Spawn Circumference", 180f, 0.01f, 360f);
				TileSpawnCircumference.setCallbackFunction += (val) => { TileSpawnCircumference.valNoCallback = val; CreateTiles(); };
				TileSpawnRadialSpacing = new JSONStorableFloat("Radial Spacing", 0.35f, 0f, 10f);
				TileSpawnRadialSpacing.setCallbackFunction += (val) => { TileSpawnRadialSpacing.valNoCallback = val; CreateTiles(); };
				
				TileHorizontalSpacing = new JSONStorableFloat("Horizontal Spacing", 0.35f, 0f, 4f);
				TileHorizontalSpacing.setCallbackFunction += (val) => { TileHorizontalSpacing.valNoCallback = val; CreateTiles(); };
				TileVerticalSpacing = new JSONStorableFloat("Vertical Spacing", 0.5f, 0f, 4f);
				TileVerticalSpacing.setCallbackFunction += (val) => { TileVerticalSpacing.valNoCallback = val; CreateTiles(); };
				
				TileLineHorizontalOffset = new JSONStorableFloat("Line Horizontal Offset", 0f, 0f, 5f, false);
				TileLineHorizontalOffset.setCallbackFunction += (val) => { TileLineHorizontalOffset.valNoCallback = val; CreateTiles(); };
				TileLineVerticalOffset = new JSONStorableFloat("Line Vertical Offset", 0f, 0f, 5f, false);
				TileLineVerticalOffset.setCallbackFunction += (val) => { TileLineVerticalOffset.valNoCallback = val; CreateTiles(); };
				TileLineRotationOffset = new JSONStorableFloat("Line Rotation Offset", 0f, 0f, 90f);
				TileLineRotationOffset.setCallbackFunction += (val) => { TileLineRotationOffset.valNoCallback = val; CreateTiles(); };

				TileScale = new JSONStorableFloat("Tile scale", 1f, 0.1f, 10f);
				TileScale.setCallbackFunction += (val) => { TileScale.valNoCallback = val; CreateTiles(); };
				
				TileRotationX = new JSONStorableFloat("Tile rotation X", 0f, 0f, 360f);
				TileRotationX.setCallbackFunction += (val) => { TileRotationX.valNoCallback = val; CreateTiles(); };
				TileRotationY = new JSONStorableFloat("Tile rotation Y", 0, 0f, 360f);
				TileRotationY.setCallbackFunction += (val) => { TileRotationY.valNoCallback = val; CreateTiles(); };
				TileRotationZ = new JSONStorableFloat("Tile rotation Z", 0, 0f, 360f);
				TileRotationZ.setCallbackFunction += (val) => { TileRotationZ.valNoCallback = val; CreateTiles(); };
				
				HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				MatAlbedoColor = new JSONStorableColor("Albedo Color", hsvcDefault, MatColorCallback);
				MatBaseColor = new JSONStorableColor("Base Color", hsvcDefault, MatColorCallback);
				
				MatAlbedoBaseColorRatio = new JSONStorableFloat("Albedo / Base Color ratio", 0f, 0f, 1f);
				MatAlbedoBaseColorRatio.setCallbackFunction += (val) => { MatAlbedoBaseColorRatio.valNoCallback = val; CreateTiles(); };
				
				MatRoughnessIntensity = new JSONStorableFloat("Roughness Intensity", 1f, 0f, 1f);
				MatRoughnessIntensity.setCallbackFunction += (val) => { MatRoughnessIntensity.valNoCallback = val; CreateTiles(); };
				
				MatNormalIntensity = new JSONStorableFloat("Normal Intensity", 1f, 0f, 1f);
				MatNormalIntensity.setCallbackFunction += (val) => { MatNormalIntensity.valNoCallback = val; CreateTiles(); };
				
				MatScale = new JSONStorableFloat("Scale", 1f, 0.1f, 20f);
				MatScale.setCallbackFunction += (val) => { MatScale.valNoCallback = val; CreateTiles(); };

				// Animation values
				EnableAnimation = new JSONStorableBool("Enable animation", false);
				EnableAnimation.setCallbackFunction += (val) => { EnableAnimation.valNoCallback = val; ToggleAnimationState(); };
				
				AnimateAlbedoColor = new JSONStorableColor("Animate Albedo Color", hsvcDefault, AnimateMatColorCallback){ isStorable=false,isRestorable=false };
				AnimateBaseColor = new JSONStorableColor("Animate Base Color", hsvcDefault, AnimateMatColorCallback){ isStorable=false,isRestorable=false };

				AnimateScale = new JSONStorableFloat("Animate scale", 0f, -10f, 10f){ isStorable=false,isRestorable=false };
				AnimateScale.setCallbackFunction += (val) => { AnimateScale.valNoCallback = val; };
				
				AnimateRotationX = new JSONStorableFloat("Animate rotation X", 0f, 0f, 360f){ isStorable=false,isRestorable=false };
				AnimateRotationX.setCallbackFunction += (val) => { AnimateRotationX.valNoCallback = val; };

				AnimateRotationY = new JSONStorableFloat("Animate rotation Y", 0f, 0f, 360f){ isStorable=false,isRestorable=false };
				AnimateRotationY.setCallbackFunction += (val) => { AnimateRotationY.valNoCallback = val; };
				
				AnimateRotationZ = new JSONStorableFloat("Animate rotation Z", 0f, 0f, 360f){ isStorable=false,isRestorable=false };
				AnimateRotationZ.setCallbackFunction += (val) => { AnimateRotationZ.valNoCallback = val; };
				
				// Help!
				_helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>3D Tiles Editor</b></size></color>\n\n" + 
					"<color=#333>" +
					"This editor allows you to customize a wall of 3D tiles.\n\n" +
					"<b>IMPORTANT:</b> Please take the time to read the animation and performances section at the end.\n\n" +
					"<b>Tiles parameters</b> All parameters to configure the tiles for static shots. Configure your base settings for the position, rotation, etc... to create your wall.\n\n" +
					"<b>Animation parameters</b> All animatable parameters for the tiles. These values are not saved and only usable as a preview or a way to see how the parameters are animated. You have to check \"Enable Animation\" (or control it using a trigger) to enable the animation.\n\n" +
					"<b>Material parameters</b> All parameters to configure the visual appearance of each tiles. Don't use these values to animate the tiles, use the animatable parameters.\n\n" +
					"<color=#000><size=30><b>Animation, performances and uses</b></size></color>\n\n" +
					"<b>First and foremost:</b> do not animate the main parameters of the tiles for a realtime use (desktop or VR). The way the generator works and allows previsualization is not made to be updated on a per frame basis during a scene or a playthrough. It could be animated if you understand perfectly how to optimize your scene or for offline rendering if your hardware is  powerful enough.\n\n" +
					"The way of animating your tiles is to control the animatable parameters. You need to check \"Enable animation\" first, then, control the color, scale or rotation through the values. They are available in the UI, but also through values readable and editable by any plugin or trigger.\n\n" +
					"<b>These animatable parameters are additive</b> meaning that they add to the current appearance and settings of the tiles. Example: if you're base tile setting for the scale is 10, animating the scale value at 5 will size your tiles to 15.\n\n" +
					"The only value that is not additive is the color. Which means that as soon as you enable animation, the color of the tiles will go back to the default one in the animation parameters. Be sure to trigger your default color with your plugin(s) to compensate for that.\n\n" +
					"<b>Uses:</b> This tile system is mostly thought for either virtual photography or music videos. But it also is pretty cool to improve or add details to simple walls... or even create abstract sculptures! Don't hesitate to be creative.\n\n" +
					"</color>"
				);
				

				RegisterBool(EnableAnimation);
				RegisterBool(EnableCollisions);
				
				RegisterColor(AnimateAlbedoColor);
				RegisterColor(AnimateBaseColor);
				RegisterFloat(AnimateScale);
				RegisterFloat(AnimateRotationX);
				RegisterFloat(AnimateRotationY);
				RegisterFloat(AnimateRotationZ);

				RegisterStringChooser(TileType);
				RegisterStringChooser(TileTypeAlt);
				RegisterStringChooser(TileMaterial);
				
				RegisterFloat(TileLineCount);
				RegisterFloat(TileRowCount);
				RegisterStringChooser(TileSpawnRadiusType);
				RegisterFloat(TileSpawnRadius);
				RegisterFloat(TileSpawnCircumference);
				RegisterFloat(TileSpawnRadialSpacing);
				RegisterFloat(TileHorizontalSpacing);
				RegisterFloat(TileVerticalSpacing);
				RegisterFloat(TileLineHorizontalOffset);
				RegisterFloat(TileLineVerticalOffset);
				RegisterFloat(TileScale);
				RegisterFloat(TileRotationX);
				RegisterFloat(TileRotationY);
				RegisterFloat(TileRotationZ);
				
				RegisterColor(MatAlbedoColor);
				RegisterColor(MatBaseColor);
				
				RegisterFloat(MatAlbedoBaseColorRatio);
				RegisterFloat(MatRoughnessIntensity);
				RegisterFloat(MatNormalIntensity);
				RegisterFloat(MatScale);
				
				// Creating the default UI that will inform you that the thing is loading
				CreateLoadingUI();
				
			}
            catch(Exception e)
            {
                SuperController.LogError("Wall3DTilesEditor - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update() {
			// If the editor is not initialized, we're gonna try to do it
			// I'm putting 3600 tries ( at 60fps it is 1min )
			if( initLoops < maxInitLoops && editorInitialized == false ) {
				initLoops++;
				// If our CUA is loading and everything is ok, we can remove the loading UI and Init the final UI
				if (InitEditor() == true)
				{
					RemoveLoadingUI();
					cuaIsValid = true; // Our CUA is initialized, we consider it as valid
					StartCoroutine("CheckCUAStatus"); // We're gonna check if the CUA is valid every 5 secs
					
					CreateTiles();
					
					// Initializing late stuff
					ToggleAnimationState();
				}
			}
			
			// Triggering an error if we have reached our limit of tries
			if( initLoops == maxInitLoops && editorInitialized == false ) {
				initLoops++; // To avoid spamming the log with the error
			}
		}
		
		private bool InitEditor() {
			// Updating our loading UI
			UpdateLoadingUI();

			if( editorInitialized == true ) return true;
			
			// Source tiles root
			TilesObjectsRoot = getChildRoot( containingAtom.reParentObject, "tiles_objects" );
			if( TilesObjectsRoot == null )
			{
				return false;
			}
			TilesObjectsRoot.gameObject.SetActive(false);
			
			// Source materials root
			TilesMaterialsRoot = getChildRoot( containingAtom.reParentObject, "tiles_materials" );
			if( TilesMaterialsRoot == null )
			{
				return false;
			}
			TilesMaterialsRoot.gameObject.SetActive(false); // By default it is, but just to be sure
			
			// The root where the tiles are gonna be created
			TilesRoot = getChildRoot( containingAtom.reParentObject, "tiles_root" );
			if( TilesRoot == null )
			{
				return false;
			}
			
			// Creating the tile types list
			TileChoices = new List<string>();
			TileChoicesTRs = new List<Transform>();
			foreach( Transform child in TilesObjectsRoot )
			{
				TileChoices.Add(child.name);
				TileChoicesTRs.Add(child);
				TileAltChoices.Add(child.name);
			}
			TileType.choices = TileChoices;
			TileTypeAlt.choices = TileAltChoices;
			
			// Creating the tile materials list
			TileMaterialChoices = new List<string>();
			TileMaterialTRs = new List<Transform>();
			foreach( Transform child in TilesMaterialsRoot )
			{
				TileMaterialChoices.Add(child.name);
				TileMaterialTRs.Add(child);
			}
			TileMaterial.choices = TileMaterialChoices;

			editorInitialized = true;
			CreateCUAUI();
			
			return true;

		}
		
		// The function creating the loading UI before everything is ready
		private void CreateLoadingUI()
		{
			_loadingInfos = new JSONStorableString("Loading Infos", "");
			UIDynamic stateinfosTextfield = CreateTextField(_loadingInfos, false);
			stateinfosTextfield.height = 450.0f;
		}

		// The function clearing the loading UI
		private void RemoveLoadingUI()
		{
			// Removing the loading text field on the left of the UI
			RemoveTextField(_loadingInfos);
		}

		// The function updating the loading UI while we're either waiting for the CUA to load, or waiting for the root object and everything to be found
		private void UpdateLoadingUI()
		{
			var loadingText = "<color=#000><b>Custom Unity Asset loading...</b></color>\n" + (initLoops/60) + "/" + (maxInitLoops/60) + "secs";
			if (initLoops == maxInitLoops)
			{
				loadingText += "\n\n<color=#333><b>UNABLE TO LOAD YOUR CUA :</b>\n";
				loadingText += "- Your CUA may not be compatible with this script\n";
				loadingText += "- Your CUA may take too long to load</color>";
				loadingText += "\n\nSelect another CUA and reload the script, or just reload the script to try again.";
			}
			_loadingInfos.val = loadingText;
		}
		
		// The function creating the UI when the CUA has been invalidated
		private void CreateCUAInvalidatedUI()
		{
			JSONStorableString _invalidatedCUA = new JSONStorableString("Invalidated CUA", "");
			UIDynamic stateinfosTextfield = CreateTextField(_invalidatedCUA, false);
			stateinfosTextfield.height = 150.0f;
			_invalidatedCUA.val = "<color=#000><b>Custom Unity Asset invalidated.</b></color>\n" +
			                      "You have changed the CUA in the <b>Asset</b> tab, the script is no longer valid. Please reload it or change it.";
		}

		// Function creating the whole CUA UI
		private void CreateCUAUI()
		{
			// Creating all the lists 
			allUIColorPickers = new List<UIDynamicColorPicker>();
			allUISliders = new List<UIDynamicSlider>();
			allUIToggles = new List<UIDynamicToggle>();
			allUITextFields = new List<UIDynamicTextField>();
			allUIButtons = new List<UIDynamicButton>();
			allUISpacers = new List<UIDynamic>();
			
			UIDynamicSlider tmpSlider;
			UIDynamicToggle tmpToggle;
			UIDynamicColorPicker tmpColor;
			
			// Static stuff
			tilesParamsUIElements = new UIDynamic[16];
			
			tmpToggle = CreateToggle(ToggleEditTilesParameters);
			allUIToggles.Add(tmpToggle);
			
			tmpToggle = CreateToggle(EnableCollisions,false);
			allUIToggles.Add(tmpToggle);
			tilesParamsUIElements[0] = (UIDynamic)tmpToggle;
			
			tmpSlider = CreateSlider(TileLineCount);
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[1] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileRowCount);
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[2] = (UIDynamic)tmpSlider;
			
			UIDynamicPopup TSRdp = CreateScrollablePopup(TileSpawnRadiusType, false);
			TSRdp.labelWidth = 180f;
			tilesParamsUIElements[3] = (UIDynamic)TSRdp;
			
			tmpSlider = CreateSlider(TileSpawnRadius);
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[4] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileSpawnCircumference);
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[5] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileSpawnRadialSpacing);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[6] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileHorizontalSpacing);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[7] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileVerticalSpacing);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[8] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileLineHorizontalOffset);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[9] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileLineVerticalOffset);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[10] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileLineRotationOffset);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[11] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileScale);
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[12] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileRotationX);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[13] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(TileRotationY);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[14] = (UIDynamic)tmpSlider;

			tmpSlider = CreateSlider(TileRotationZ);
			tmpSlider.valueFormat = "F3";
			allUISliders.Add(tmpSlider);
			tilesParamsUIElements[15] = (UIDynamic)tmpSlider;
			
			// Animation
			tilesAnimUIElements = new UIDynamic[7];
			
			tmpToggle = CreateToggle(ToggleAnimateTilesParameters);
			allUIToggles.Add(tmpToggle);
			
			tmpToggle = CreateToggle(EnableAnimation);
			tilesAnimUIElements[0] = (UIDynamic)tmpToggle;
			
			tmpColor = CreateColorPicker(AnimateAlbedoColor, false);
			allUIColorPickers.Add(tmpColor);
			tilesAnimUIElements[1] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(AnimateBaseColor, false);
			allUIColorPickers.Add(tmpColor);
			tilesAnimUIElements[2] = (UIDynamic)tmpColor;
			
			tmpSlider = CreateSlider(AnimateScale);
			allUISliders.Add(tmpSlider);
			tilesAnimUIElements[3] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(AnimateRotationX);
			allUISliders.Add(tmpSlider);
			tilesAnimUIElements[4] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(AnimateRotationY);
			allUISliders.Add(tmpSlider);
			tilesAnimUIElements[5] = (UIDynamic)tmpSlider;

			tmpSlider = CreateSlider(AnimateRotationZ);
			allUISliders.Add(tmpSlider);
			tilesAnimUIElements[6] = (UIDynamic)tmpSlider;
			
			// Material
			tilesMaterialUIElements = new UIDynamic[9];
			
			tmpToggle = CreateToggle(ToggleMaterialParameters,true);
			allUIToggles.Add(tmpToggle);
			
			UIDynamicPopup TTdp = CreateScrollablePopup(TileType, true);
			TTdp.labelWidth = 180f;
			tilesMaterialUIElements[0] = (UIDynamic)TTdp;
			
			UIDynamicPopup TTAdp = CreateScrollablePopup(TileTypeAlt, true);
			TTAdp.labelWidth = 180f;
			tilesMaterialUIElements[1] = (UIDynamic)TTAdp;
			
			UIDynamicPopup TMdp = CreateScrollablePopup(TileMaterial, true);
			TMdp.labelWidth = 180f;
			tilesMaterialUIElements[2] = (UIDynamic)TMdp;
			
			tmpColor = CreateColorPicker(MatAlbedoColor, true);
			allUIColorPickers.Add(tmpColor);
			tilesMaterialUIElements[3] = (UIDynamic)tmpColor;
			
			tmpColor = CreateColorPicker(MatBaseColor, true);
			allUIColorPickers.Add(tmpColor);
			tilesMaterialUIElements[4] = (UIDynamic)tmpColor;
			
			tmpSlider = CreateSlider(MatAlbedoBaseColorRatio, true);
			allUISliders.Add(tmpSlider);
			tilesMaterialUIElements[5] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(MatRoughnessIntensity, true);
			allUISliders.Add(tmpSlider);
			tilesMaterialUIElements[6] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(MatNormalIntensity, true);
			allUISliders.Add(tmpSlider);
			tilesMaterialUIElements[7] = (UIDynamic)tmpSlider;
			
			tmpSlider = CreateSlider(MatScale, true);
			allUISliders.Add(tmpSlider);
			tilesMaterialUIElements[8] = (UIDynamic)tmpSlider;
			
			HelpUIElement = CreateTextField(_helpText, true);
			HelpUIElement.height = 900.0f;
			
			ToggleEditTilesParametersCallback(true);
			ToggleEditRadiusElements();
			ToggleAnimateTilesParametersCallback(false);
			ToggleMaterialParametersCallback(false);
		}
		
		// Function clearing the whole CUA UI
		private void ClearCUAUI()
		{		
			foreach( UIDynamicColorPicker uiEl in allUIColorPickers ) {
				RemoveColorPicker(uiEl);
			}
			
			foreach( UIDynamicSlider uiEl in allUISliders ) {
				RemoveSlider(uiEl);
			}
			
			foreach( UIDynamicToggle uiEl in allUIToggles ) {
				RemoveToggle(uiEl);
			}
			
			foreach( UIDynamicTextField uiEl in allUITextFields ) {
				RemoveTextField(uiEl);
			}
			
			foreach( UIDynamic uiEl in allUISpacers ) {
				RemoveSpacer(uiEl);
			}
			
			foreach( UIDynamicButton uiEl in allUIButtons ) {
				RemoveButton(uiEl);
			}
			
			RemoveTextField( _helpText );
		}
			
		// **************************
		// Functions
		// **************************
		private void CreateTiles() 
		{
			if( TilesRoot == null ) return;
			
			// Getting the tile
			Transform tmp = TileChoicesTRs[ TileChoices.IndexOf( TileType.val ) ];
			tmp.gameObject.SetActive(true);
			
			// If we need an alt tile
			Transform tmpAlt = null;
			if (TileTypeAlt.val != "None")
			{
				tmpAlt = TileChoicesTRs[TileAltChoices.IndexOf(TileTypeAlt.val) - 1];
				tmpAlt.gameObject.SetActive(true);
			}

			// Setting the material for the tile
			UpdateTileMaterial( tmp );
			if (tmpAlt != null) UpdateTileMaterial(tmpAlt);
			
			// Cleaning the parent which contains all tiles
			foreach (Transform child in TilesRoot)
			{
				Destroy(child.gameObject);
			}
			
			if( tmp == null ) return;
			
			GameObject newTile = null;
			GameObject clone = tmp.gameObject;
			GameObject cloneAlt = null;
			if (tmpAlt != null) cloneAlt = tmpAlt.gameObject;
			
			float hOffset = 0f;
			float vOffset = 0f;
			float pOffset = 0f; // "Plane" offset for each "radial" iteration
			float lineHorOffset = 0f;
			float lineVertOffset = 0f;
			float lineRotOffset = 0f;
			CurrentTilesList = new List<Transform>();
			Vector3 tilePos = new Vector3(0,0,0);
			Quaternion tileRot;
			GameObject currentLine;
			
			// Enabling collisions if needed
			MeshCollider tmpCol = tmp.GetComponent<MeshCollider>();
			if( tmpCol != null ) tmpCol.enabled = EnableCollisions.val;
			if ( tmpAlt != null )
			{
				MeshCollider tmpColAlt = tmpAlt.GetComponent<MeshCollider>();
				if( tmpColAlt != null ) tmpColAlt.enabled = EnableCollisions.val;
			}
			
			for( var j = 0; j < TileLineCount.val; j++ ) {
				
				currentLine = new GameObject("tile_line_" + j);
				currentLine.transform.SetParent(TilesRoot);
				currentLine.transform.localPosition = new Vector3(lineHorOffset,vOffset,lineVertOffset);
				currentLine.transform.localRotation = Quaternion.Euler(0f, lineRotOffset, 0f);
				
				for( var i = 0; i < TileRowCount.val; i++ ) {
					
					if( TileSpawnRadius.val > 0f && TileSpawnRadiusType.val == "Radial Plane" ) {
						var rCount = TileSpawnCircumference.val < 360 ? TileRowCount.val - 1 : TileRowCount.val;
						float angle = i * ( TileSpawnCircumference.val * Mathf.PI / 180 ) / rCount;
						
						tilePos = new Vector3( Mathf.Cos(angle)*(TileSpawnRadius.val + pOffset), Mathf.Sin(angle)*(TileSpawnRadius.val+pOffset), 0f);
						tileRot = Quaternion.Euler(TileRotationX.val, TileRotationY.val,  ((TileSpawnCircumference.val / rCount) * i - 90f) + TileRotationZ.val);
					} else if( TileSpawnRadius.val > 0f && TileSpawnRadiusType.val == "Radial 3D" ) {
						var rCount = TileSpawnCircumference.val < 360 ? TileRowCount.val - 1 : TileRowCount.val;
						float angle = i * ( TileSpawnCircumference.val * Mathf.PI / 180 ) / rCount;
						
						tilePos = new Vector3( Mathf.Cos(angle)*TileSpawnRadius.val, 0f, Mathf.Sin(angle)*TileSpawnRadius.val);						
						tileRot = Quaternion.Euler(TileRotationX.val, - ((TileSpawnCircumference.val / rCount) * i - 90f) + TileRotationY.val, TileRotationZ.val);
					} else {
						tilePos = new Vector3(hOffset,0f,0f);
						tileRot = Quaternion.Euler(TileRotationX.val, TileRotationY.val, TileRotationZ.val);
					}

					newTile = clone;
					if (cloneAlt != null && i%2!=0) newTile = cloneAlt;
					
					newTile = Instantiate(newTile, currentLine.transform);
					newTile.name = "tile_"+j+"_"+i;
					newTile.transform.localPosition = tilePos;
					newTile.transform.localScale = new Vector3( TileScale.val,TileScale.val,TileScale.val);
					newTile.transform.localRotation = tileRot;

					hOffset += -TileHorizontalSpacing.val;
				}
				
				hOffset = 0;
				if( TileSpawnRadiusType.val == "Flat" || TileSpawnRadiusType.val == "Radial 3D" ) vOffset += TileVerticalSpacing.val;
				if( TileSpawnRadiusType.val == "Radial Plane" ) pOffset += TileSpawnRadialSpacing.val;
				lineHorOffset = (j%2==0) ? TileLineHorizontalOffset.val : 0;
				lineVertOffset = (j%2==0) ? TileLineVerticalOffset.val : 0;
				lineRotOffset = (j%2==0) ? TileLineRotationOffset.val : 0;
			}
			
			// When everything is created, converting the TR list to an array (faster iteration)
			CurrentTilesArray = CurrentTilesList.ToArray();
		}
		
		private void UpdateTileMaterial( Transform tileTR )
		{
            try
            {
				// Grabbing the Tile renderer
				Renderer tmpTileRend = tileTR.GetComponent<Renderer>();
				
				// Grabbing the source material
				Transform tmpMatTr = TileMaterialTRs[ TileMaterialChoices.IndexOf( TileMaterial.val ) ];
				Renderer tmpMatRend = tmpMatTr.GetComponent<Renderer>();
				Material tmpMat = tmpMatRend.material;
						
				// Changing Albedo color
				Color newColor = new Color(MatAlbedoColor.colorPicker.currentColor.r, MatAlbedoColor.colorPicker.currentColor.g, MatAlbedoColor.colorPicker.currentColor.b, 1f);
				tmpMat.SetColor("_AlbedoColor",newColor);
				
				// Changing Base color
				newColor = new Color(MatBaseColor.colorPicker.currentColor.r, MatBaseColor.colorPicker.currentColor.g, MatBaseColor.colorPicker.currentColor.b, 1f);
				tmpMat.SetColor("_BaseColor",newColor);
				
				// Changing albedo / base color ratio
				tmpMat.SetFloat("_AlbedoBaseColorRatio",MatAlbedoBaseColorRatio.val);
				
				// Changing roughness intensity
				tmpMat.SetFloat("_RoughnessIntensity",MatRoughnessIntensity.val);
				
				// Changing normal intensity
				tmpMat.SetFloat("_NormalIntensity",MatNormalIntensity.val);
				
				// Changing the material on the tile
				tmpTileRend.material = tmpMat;

				tmpMat.SetTextureScale("_Albedo", new Vector2(MatScale.val, MatScale.val));
				tmpMat.SetTextureScale("_Normal", new Vector2(MatScale.val, MatScale.val));
				tmpMat.SetTextureScale("_Mask", new Vector2(MatScale.val, MatScale.val));
            }
            catch(Exception e)
            {
                SuperController.LogError("Wall3DTilesEditor - Exception caught while updating the material: " + e);
            }
		
		}
		
		protected void ToggleAnimationState() {
			if( animRoutineRunning == false ) {
				StartCoroutine("AnimateTiles");
			}
		}
			
		protected void MatColorCallback(JSONStorableColor color) {
			// Not using the value, we're gonna get every JSONStorable values directly
			CreateTiles();
		}
		
		protected void AnimateMatColorCallback(JSONStorableColor color) {
			// Nothing
		}
		
		protected void ToggleEditTilesParametersCallback(bool enabled) {
			foreach( UIDynamic uidElem in tilesParamsUIElements ) {
				if( enabled == true ) {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = false;
				} else {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = true;
				}
			}
		}
		
		protected void ToggleEditRadiusElements()
		{
			if (tilesParamsUIElements == null) return;
			if (TileSpawnRadiusType == null || TileSpawnRadiusType.val == null) return;
			if( TileSpawnRadiusType.val != "Flat" ) {
				tilesParamsUIElements[4].GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
				tilesParamsUIElements[4].GetComponent<LayoutElement>().ignoreLayout = false;
				tilesParamsUIElements[5].GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
				tilesParamsUIElements[5].GetComponent<LayoutElement>().ignoreLayout = false;

				if (TileSpawnRadiusType.val == "Radial Plane")
				{
					tilesParamsUIElements[6].GetComponent<LayoutElement>().transform.localScale = new Vector3(1, 1, 1);
					tilesParamsUIElements[6].GetComponent<LayoutElement>().ignoreLayout = false;
					
					tilesParamsUIElements[7].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					tilesParamsUIElements[7].GetComponent<LayoutElement>().ignoreLayout = true;
					tilesParamsUIElements[8].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					tilesParamsUIElements[8].GetComponent<LayoutElement>().ignoreLayout = true;
				}
				else
				{
					tilesParamsUIElements[8].GetComponent<LayoutElement>().transform.localScale = new Vector3(1, 1, 1);
					tilesParamsUIElements[8].GetComponent<LayoutElement>().ignoreLayout = false;
					
					tilesParamsUIElements[6].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					tilesParamsUIElements[6].GetComponent<LayoutElement>().ignoreLayout = true;					
				}
			} else {
				tilesParamsUIElements[4].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
				tilesParamsUIElements[4].GetComponent<LayoutElement>().ignoreLayout = true;
				tilesParamsUIElements[5].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
				tilesParamsUIElements[5].GetComponent<LayoutElement>().ignoreLayout = true;
				tilesParamsUIElements[6].GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
				tilesParamsUIElements[6].GetComponent<LayoutElement>().ignoreLayout = true;
				
				tilesParamsUIElements[7].GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
				tilesParamsUIElements[7].GetComponent<LayoutElement>().ignoreLayout = false;
				tilesParamsUIElements[8].GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
				tilesParamsUIElements[8].GetComponent<LayoutElement>().ignoreLayout = false;
			}
		}
		
		protected void ToggleAnimateTilesParametersCallback(bool enabled) {
			foreach( UIDynamic uidElem in tilesAnimUIElements ) {
				if( enabled == true ) {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = false;
				} else {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = true;
				}
			}
		}
		
		protected void ToggleMaterialParametersCallback(bool enabled) {
			foreach( UIDynamic uidElem in tilesMaterialUIElements ) {
				if( enabled == true ) {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = false;
				} else {
					uidElem.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
					uidElem.GetComponent<LayoutElement>().ignoreLayout = true;
				}
			}
			
			if( enabled == true ) {			
				HelpUIElement.GetComponent<LayoutElement>().transform.localScale = new Vector3(0,0,0);
				HelpUIElement.GetComponent<LayoutElement>().ignoreLayout = true;
			} else {
				HelpUIElement.GetComponent<LayoutElement>().transform.localScale = new Vector3(1,1,1);
				HelpUIElement.GetComponent<LayoutElement>().ignoreLayout = false;
			}
		}

		
		// **************************
		// COROUTINES
		// **************************
		
		// Routine checking every 5secs if our CUA is still valid, this will invalidate the UI if the user change the CUA in the asset tab
		private IEnumerator CheckCUAStatus()
		{
			while (cuaIsValid)
			{
				// Our object root is null, it means our CUA is invalid
				if (TilesObjectsRoot == null)
				{
					// We invalidate it so that the coroutine stops by itself, remove the UI and show the UI to alert the user
					cuaIsValid = false;
					ClearCUAUI();
					CreateCUAInvalidatedUI();
				}
				yield return new WaitForSeconds(5.0f);
			}
		}
		
		// Animation routine
		private IEnumerator AnimateTiles()
		{
			animRoutineRunning = true;
			while ( EnableAnimation.val == true && TilesRoot != null && TilesRoot.childCount > 0 )
			{
				float rotX = 0;
				float rotY = 0;
				float scale = 0;
				Vector3 tilePos = new Vector3(0,0,0);
				Quaternion tileRot;
				for (var m = 0; m < TilesRoot.childCount; m++ ) {
					Transform tmpLine = TilesRoot.GetChild(m);
					for (var n = 0; n < tmpLine.childCount; n++ ) {
						Transform tmpTile = tmpLine.GetChild(n);

						scale = TileScale.val + AnimateScale.val;
						if( TileSpawnRadius.val > 0f && TileSpawnRadiusType.val == "Radial Plane" ) {
							var rCount = TileSpawnCircumference.val < 360 ? TileRowCount.val - 1 : TileRowCount.val;
							float angle = n * ( TileSpawnCircumference.val * Mathf.PI / 180 ) / rCount;
							tileRot = Quaternion.Euler(TileRotationX.val + AnimateRotationX.val,  TileRotationY.val + AnimateRotationY.val, ((TileSpawnCircumference.val / rCount) * n - 90f) + TileRotationZ.val + AnimateRotationZ.val);
						} else if( TileSpawnRadius.val > 0f && TileSpawnRadiusType.val == "Radial 3D" ) {
							var rCount = TileSpawnCircumference.val < 360 ? TileRowCount.val - 1 : TileRowCount.val;
							float angle = n * ( TileSpawnCircumference.val * Mathf.PI / 180 ) / rCount;
							tileRot = Quaternion.Euler(TileRotationX.val + AnimateRotationX.val, - ((TileSpawnCircumference.val / rCount) * n - 90f) + TileRotationY.val + AnimateRotationY.val, TileRotationZ.val + AnimateRotationZ.val);
						} else {
							tileRot = Quaternion.Euler(TileRotationX.val + AnimateRotationX.val, TileRotationY.val + AnimateRotationY.val, TileRotationZ.val + AnimateRotationZ.val);
						}
						
						tmpTile.transform.localScale = new Vector3( scale, scale, scale);
						tmpTile.transform.localRotation = tileRot;
						
						// Grabbing the Tile renderer
						Renderer tmpTileRend = tmpTile.GetComponent<Renderer>();				
						Material tmpMat = tmpTileRend.material;

						// Changing Albedo color
						Color newColor = new Color(AnimateAlbedoColor.colorPicker.currentColor.r, AnimateAlbedoColor.colorPicker.currentColor.g, AnimateAlbedoColor.colorPicker.currentColor.b, 1f);
						tmpMat.SetColor("_AlbedoColor",newColor);
						
						// Changing Base color
						newColor = new Color(AnimateBaseColor.colorPicker.currentColor.r, AnimateBaseColor.colorPicker.currentColor.g, AnimateBaseColor.colorPicker.currentColor.b, 1f);
						tmpMat.SetColor("_BaseColor",newColor);
					}
				}
				yield return null;
			}
			animRoutineRunning = false;
		}

		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private JSONStorableString createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescString;
		}
				
		private Transform getChildRoot( Transform parent, string searchName ) {
			foreach( Transform child in parent )
			{
				if( child.name == searchName ) {
					return child;
				} else {
					Transform childSearch = getChildRoot(child,searchName);
					if (childSearch != null) return childSearch;
				}
			}
			
			return null;
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified native VAM assets

		}

	}
}